import sys

from .fancy_number import FancyInt
from .name_fmt import RandomizedNameFmt


class NamedNumber(FancyInt):
    def __new__(cls, i: int | str | None = None, fmt: str = "%adjective% %animal%",
                fmt_type: type = RandomizedNameFmt, groups: dict | None = None, rng: None = None,
                random_seed: int | None = 12345, options: dict | None = None,  byteorder: str = sys.byteorder,
                encrypt=None, decrypt=None, **group_kwargs):

        name_fmt = fmt_type(fmt=fmt, groups=groups, rng=rng, random_seed=random_seed, options=options,
                            byteorder=byteorder, encrypt=encrypt, decrypt=decrypt, **group_kwargs) if isinstance(fmt, str) else fmt

        # if no value is specified, generate a random value
        if i is None or i == "random":
            i = name_fmt.random_number()

        # if a string is specified, convert it to int
        if isinstance(i, str):
            i = name_fmt.int_from_name(i)

        if isinstance(i, (list, tuple, set)):
            return type(i)(cls(_i, fmt=fmt, fmt_type=fmt_type, groups=groups, rng=rng, random_seed=random_seed,
                               options=options, byteorder=byteorder, encrypt=encrypt, decrypt=decrypt, **group_kwargs) for _i in i)

        if isinstance(i, slice):
            start = i.start if i.start is not None else 0
            stop = i.stop if i.stop is not None else name_fmt.max_number + 1
            step = i.step if i.step is not None else 1
            i = range(start, stop, step)

        if isinstance(i, range):
            return [cls(_i, fmt=fmt, fmt_type=fmt_type, groups=groups, rng=rng, random_seed=random_seed,
                        options=options, byteorder=byteorder, encrypt=encrypt, decrypt=decrypt, **group_kwargs) for _i in i]

        if not isinstance(i, int):
            try:
                i = int(i)
            except:
                raise Exception(f"invalid input type: {type(i)}")

        # if not in valid range, just return int instead
        if isinstance(i, int) and (i < 0 or i > name_fmt.max_number):
            return i

        name = name_fmt.name_from_int(i)
        strings = name_fmt.strings_from_name(name)
        indices = name_fmt.indices_from_name(strings)

        self = int.__new__(cls, i)
        self.fmt = name_fmt
        self.name = name
        self.strings = strings
        self.indices = indices
        return self

    def __repr__(self):
        return f'<{self.name} ({super().__repr__()})>'

    def __str__(self):
        return self.name

    def math_result(self, r):
        if isinstance(r, int) and 0 <= r <= self.fmt.max_number:
            r = self.fmt(r)
        return r

    def __contains__(self, item):
        return item in self.name

    def __iter__(self):
        for i in range(len(self.indices)):
            yield self.strings[i], self.indices[i]

    def __getitem__(self, item):
        return self.name[item]

    def __setitem__(self, key, value):
        raise NotImplementedError("cannot set item on NamedNumber")

    def __eq__(self, other):
        r = super().__eq__(other)
        return r or isinstance(other, str) and other == self.name


if __name__ == "__main__":
    p = NamedNumber(0)



import argparse
import json
import sys

from .named_number import NamedNumber
from .name_fmt import RandomizedNameFmt, NameFmt, IncrementingNameFmt


def main():
    parser = argparse.ArgumentParser(description='convert an integer to a number')
    parser.add_argument('integer', metavar='i', nargs='*', help='the integer to convert')
    parser.add_argument('--fmt', metavar='f', type=str, default="%adjective% %animal%",
                        help='the format to use for the string', required=False)
    parser.add_argument('--shuffle', action='store_true')
    parser.add_argument('--inc', dest='shuffle', action='store_false')
    parser.set_defaults(shuffle=True)
    parser.add_argument('--seed', metavar="s", type=int, default=12345, required=False)
    parser.add_argument('--byteorder', metavar="b", type=str, default=sys.byteorder, required=False)

    args, unknown = parser.parse_known_args()

    kwargs = {}
    for arg in unknown:
        if arg.startswith(("-", "--")):
            name, *values = arg.split('=')
            value = "=".join(values).replace(",", '","').replace("[",'["').replace("]",'"]')
            value = json.loads(value)
            kwargs[name[2:]] = value

    def check(v):
        if v is None:
            return v
        if v.isnumeric():
            return int(v)
        return slice(*[int(_v) if _v else None for _v in v.split(":")])

    i = [check(v) for v in args.integer]
    if len(i) == 0:
        i = None
    elif len(i) == 1:
        i = i[0]

    nn = NamedNumber(i, fmt=args.fmt, fmt_type=RandomizedNameFmt if args.shuffle else IncrementingNameFmt,
                     random_seed=args.seed, byteorder=args.byteorder, **kwargs)

    if isinstance(nn, (list, tuple, set)):
        for _nn in nn:
            print(_nn)
    else:
        print(nn)


if __name__ == "__main__":
    main()



class FancyNumber(object):
    """fancy class which can be used to subclass a number so that the result of mathmatecal operations
    return an instance of the fancy number object"""

    def math_result(self, r):
        """cast the result of a mathematical operation into"""
        try:
            return type(self)(r)
        except:
            return r

    def __add__(self, *a, **kw):
        return self.math_result(super().__add__(*a, **kw))

    def __radd__(self, *a, **kw):
        return self.math_result(super().__radd__(*a, **kw))

    def __sub__(self, *a, **kw):
        return self.math_result(super().__sub__(*a, **kw))

    def __rsub__(self, *a, **kw):
        return self.math_result(super().__rsub__(*a, **kw))

    def __mul__(self, *a, **kw):
        return self.math_result(super().__mul__(*a, **kw))

    def __rmul__(self, *a, **kw):
        return self.math_result(super().__rmul__(*a, **kw))

    def __matmul__(self, *a, **kw):
        return self.math_result(super().__matmul__(*a, **kw))

    def __rmatmul__(self, *a, **kw):
        return self.math_result(super().__rmatmul__(*a, **kw))

    def __rdiv__(self, *a, **kw):
        return self.math_result(super().__rdiv__(*a, **kw))

    def __truediv__(self, *a, **kw):
        return self.math_result(super().__truediv__(*a, **kw))

    def __rtruediv__(self, *a, **kw):
        return self.math_result(super().__rtruediv__(*a, **kw))

    def __floordiv__(self, *a, **kw):
        return self.math_result(super().__floordiv__(*a, **kw))

    def __rfloordiv__(self, *a, **kw):
        return self.math_result(super().__rfloordiv__(*a, **kw))

    def __divmod__(self, *a, **kw):
        return self.math_result(super().__divmod__(*a, **kw))

    def __rdivmod__(self, *a, **kw):
        return self.math_result(super().__rdivmod__(*a, **kw))

    def __mod__(self, *a, **kw):
        return self.math_result(super().__mod__(*a, **kw))

    def __rmod__(self, *a, **kw):
        return self.math_result(super().__rmod__(*a, **kw))

    def __rshift__(self, *a, **kw):
        return self.math_result(super().__rshift__(*a, **kw))

    def __rrshift__(self, *a, **kw):
        return self.math_result(super().__rrshift__(*a, **kw))

    def __lshift__(self, *a, **kw):
        return self.math_result(super().__lshift__(*a, **kw))

    def __rlshift__(self, *a, **kw):
        return self.math_result(super().__rlshift__(*a, **kw))

    def __pow__(self, *a, **kw):
        return self.math_result(super().__pow__(*a, **kw))

    def __rpow__(self, *a, **kw):
        return self.math_result(super().__rpow__(*a, **kw))

    def __and__(self, *a, **kw):
        return self.math_result(super().__and__(*a, **kw))

    def __rand__(self, *a, **kw):
        return self.math_result(super().__rand__(*a, **kw))

    def __or__(self, *a, **kw):
        return self.math_result(super().__or__(*a, **kw))

    def __ror__(self, *a, **kw):
        return self.math_result(super().__ror__(*a, **kw))

    def __xor__(self, *a, **kw):
        return self.math_result(super().__xor__(*a, **kw))

    def __rxor__(self, *a, **kw):
        return self.math_result(super().__rxor__(*a, **kw))

    def __abs__(self, *a, **kw):
        return self.math_result(super().__abs__(*a, **kw))

    def __invert__(self, *a, **kw):
        return self.math_result(super().__invert__(*a, **kw))

    def __pos__(self, *a, **kw):
        return self.math_result(super().__pos__(*a, **kw))

    def __neg__(self, *a, **kw):
        return self.math_result(super().__neg__(*a, **kw))

    def __round__(self, *a, **kw):
        return self.math_result(super().__round__(*a, **kw))

    def __trunc__(self, *a, **kw):
        return self.math_result(super().__trunc__(*a, **kw))


class FancyInt(FancyNumber, int):
    """a fancy integer which can be extended and will remain a FancyInt even when math is done to it"""
    @classmethod
    def from_bytes(cls, *a, **kw) -> int:
        return cls(int.from_bytes(*a, **kw))


class FancyFloat(FancyNumber, float):
    """a fancyfloat which can be extended and will remain a FancyFloat even when math is done to it"""
    @staticmethod
    def fromhex(*args, **kwargs):
        return FancyFloat(float.fromhex(*args, *kwargs))

    def __ceil__(self, *a, **kw):
        return self.math_result(super().__ceil__(*a, **kw))


if __name__ == "__main__":
    import os

    class MyMagicNumber(FancyInt):
        """a fun integer which prints itself as an additional expression"""
        def __repr__(self):
            rand = int(int.from_bytes(os.urandom(32), "little")/(1 << (32*8))*self)
            return f"{int(self -rand)} + {rand}"

    s = MyMagicNumber(50)
    for _ in range(10):
        print(f"my number is equal to {s}")



import re
import sys
import os
import math
import time
from typing import List, Tuple, Any

try:
    # trying to avoid requiring numpy (or any external packages)
    # although it is required for the RandomizedNameFmt class
    import numpy as np
except:
    np = None


from .options import Options, Optionset, Charset


class NameFmt(object):
    name_pattern = "[a-zA-Z0-9]+[a-zA-Z0-9_-]*?"
    pattern = rf"%({name_pattern}#?\d*)%"
    """regular expression pattern used to match group names in format"""

    @staticmethod
    def sub_list(pattern: str, values: list, s: str) -> str:
        """substitute values in order for each pattern match in some string"""
        i = iter(values)
        return re.sub(pattern, lambda *a: str(next(i)), s)

    @staticmethod
    def prod(arr: List[int]) -> int:
        """get the product of integers in a list"""
        mx = arr[-1]
        for b in reversed(arr[:-1]):
            mx *= b
        return mx

    @classmethod
    def list_of_components(cls, i: int, bases: list[int] | int | None = None, base: int | None = 10,
                           byteorder: str = sys.byteorder):
        """convert an integer to a list of components using a list of bases"""
        if bases is None:
            bases = base
        if isinstance(bases, int):
            bases = math.ceil(math.log(i, bases))
        else:
            mx = cls.prod(bases) - 1
            assert 0 <= i <= mx, f"value out of range [0, {mx}]"

        ordered_bases = cls.sort_by_byteorder(bases, byteorder)

        values = []
        for base in ordered_bases:
            v = i % base
            i //= base
            values.insert(0, v)
        return values

    @classmethod
    def sort_by_byteorder(cls, values: List[int | Any], byteorder: str | List[int | float] = sys.byteorder) -> list:
        assert byteorder in ["little", "big"] or isinstance(byteorder, (list, tuple)) and len(byteorder) == len(values), \
            "byte order must be 'little', 'big', or a list of values"
        if isinstance(byteorder, str):
            byteorder = byteorder.lower()
            assert byteorder in ["little", "big"], "byte order must be 'little' or 'big'"
            ordered_values = values if byteorder == "big" else reversed(values)
        elif isinstance(byteorder, (list, tuple)) and len(byteorder) == len(values):
            ordered_values = [base for _, base in sorted(zip(byteorder, values))]
        else:
            raise Exception("invalid byteorder")
        return ordered_values

    @classmethod
    def interpret_format(cls, fmt: str, options: Options | None = None) -> \
            Tuple[str, str, List[str], List[Optionset | str | List[str]], List[int], int]:
        """interpret a format string """
        if isinstance(fmt, cls):
            return fmt.fmt, fmt.match_pattern, fmt.group_names, fmt.groups, fmt.bases, fmt.max_number

        group_strs = re.findall(cls.pattern, fmt)
        group_names = []
        """list of names of all the groups used in the format"""
        fmts = []
        for group_str in group_strs:
            group_name, *a = group_str.split("#")
            group_names.append(group_name)
            n = int(a[0]) if a else 1
            group_names += (n - 1) * [group_name]
            fmts.append(f"%{group_name}%" * n)

        # safe_fmt = re.escape
        # safe_fmt = fmt  # TODO escape the input format string so that it can include . and other special characters
        fmt = cls.sub_list(cls.pattern, fmts, fmt)

        if options is None:
            options = Options()
        groups = [options[group_name] for group_name in group_names]
        bases = [len(group) for group in groups]

        group_match_patterns = [f"([{group}])" if isinstance(group, (Charset, str)) else f"({cls.name_pattern})" for group in groups]
        match_pattern = cls.sub_list(cls.pattern, group_match_patterns, fmt)
        max_number = cls.prod(bases) - 1
        return fmt, match_pattern, group_names, groups, bases, max_number

    def __init__(self, fmt: str = "%adjective% %animal%", groups: dict | None = None, rng: None = None,
                 random_seed: int | None = 12345, options: Options | None = None,  byteorder: str = sys.byteorder,
                 encrypt=None, decrypt=None, **group_kwargs):
        if options is None:
            options = Options()
        self.options = options
        if groups is not None:
            self.options.update(groups)
        self.options.update(group_kwargs)

        if encrypt is not None:
            self.encrypt = encrypt
        if decrypt is not None:
            self.decrypt = decrypt

        self.original_fmt = fmt
        self.byteorder = byteorder
        self.fmt, self.match_pattern, self.group_names, self.groups, self.bases, self.max_number = self.interpret_format(fmt, options)
        self.random_seed = random_seed
        self.rng = np.random.default_rng(random_seed) if np is not None and rng is None  and random_seed is not None else rng
        self.init_cipher()

    def name_from_int(self, i: int) -> str:
        i = self.encrypt(i)
        name = self._name_from_int(i)
        return self.reformat(name)

    def _name_from_int(self, i: int) -> str:
        assert 0 <= i <= self.max_number, "integer out of range"
        indices = self.list_of_components(i, self.bases, byteorder=self.byteorder)
        values = [self.groups[i][ind] for i, ind in enumerate(indices)]
        return self.sub_list(self.pattern, values, self.fmt)

    def strings_from_name(self, name: str) -> List[str]:
        name = self.deformat(name)
        return self._strings_from_name(name)

    def _strings_from_name(self, name: str) -> List[str]:
        values = re.fullmatch(self.match_pattern, name).groups()
        return list(values)

    def indices_from_name(self, name: str | List[str]) -> List[int]:
        name = self.deformat(name) if isinstance(name, str) else [self.deformat(n) for n in name]
        return self._indices_from_name(name)

    def _indices_from_name(self, name: str | List[str]) -> List[int]:
        if isinstance(name, str):
            values = self._strings_from_name(name)
        else:
            values = name
        indices = [group.index(values[i]) for i, group in enumerate(self.groups)]
        return indices

    def int_from_indices(self, indices: List[int]) -> int:
        v = self._int_from_indices(indices)
        return self.decrypt(v)

    def _int_from_indices(self, indices: List[int]) -> int:
        v = indices[-1]
        for i, _v in enumerate(reversed(indices[:-1])):
            v += _v * self.prod(self.bases[-i-1:])
        return v

    def int_from_name(self, name: str):
        return self.int_from_indices(self.indices_from_name(name))

    def _int_from_name(self, name: str):
        return self._int_from_indices(self._indices_from_name(name))

    def random_number(self) -> int:
        r = int.from_bytes(os.urandom(int(self.max_number.bit_length()/8)), self.byteorder) % (self.max_number + 1)
        return r

    def random_named_number(self) -> int:
        from named_number import NamedNumber
        return NamedNumber(self.random_number(), fmt=self)

    def named_number(self, i: int | None = None, **kw) -> int:
        from named_number import NamedNumber
        if i is None:
            i = self.random_number()
        return NamedNumber(i, fmt=self, **kw)

    def __len__(self) -> int:
        return self.max_number

    def __repr__(self) -> str:
        return f"<NameFmt('{self.original_fmt}')>"

    def __call__(self, i: int | None = None, **kw) -> int:
        return self.named_number(i, **kw)

    def range(self, *a):
        return [self(i) for i in range(*a)]

    def __getitem__(self, item):
        return self(item)

    def init_cipher(self):
        pass

    def encrypt(self, i):
        return i

    def decrypt(self, i):
        return i

    def deformat(self, name):
        return name

    def reformat(self, name):
        return name

    def plot_encryption(self, x=100, require_reversible=True):
        import matplotlib.pyplot as plt

        if isinstance(x, int):
            x = range(x)

        if isinstance(x, slice):
            start = x.start if x.start is None else 0
            stop = x.stop if x.stop is not None else self.max_number + 1
            step = x.step if x.step is not None else 1
            x = range(start, stop, step)

        y = []
        for _x in x:
            _y = self.encrypt(_x)
            if require_reversible:
                assert self.decrypt(_y) == _x, "encryption not reversible"
            y.append(_y)
        plt.scatter(x, y)


class IncrementingNameFmt(NameFmt):
    pass


class RandomizedNameFmt(NameFmt):
    max_size_allowed = 1 << 23
    mappings = {}

    def init_cipher(self):
        if self.max_number not in self.mappings:
            n = self.max_number + 1
            if n > self.max_size_allowed:
                raise Exception(f"RandomizedNameFmt does not support sets greater than {self.max_size_allowed} ({n})")
            # print(f"shuffling array of size {n}")
            t0 = time.time()
            m = self.rng.permutation(self.max_number + 1)
            e = time.time() - t0
            # print(f"{e=}")
            self.mappings[self.max_number] = m
        self.mapping = self.mappings[self.max_number]

    def encrypt(self, i):
        return self.mapping[i]

    def decrypt(self, i):
        return np.argmax(self.mapping==i)

    def plot_performance(self, bit_length_range=range(28)):
        import matplotlib.pyplot as plt

        def shuffle_time(n):
            t0 = time.time()
            self.rng.permutation(n)
            return time.time() - t0

        x = list(bit_length_range)
        y = [shuffle_time(1<<_x) for _x in x]
        plt.plot(y)


if __name__ == "__main__":
    import matplotlib
    matplotlib.use('TkAgg')

    fmt = RandomizedNameFmt("%adjective% %color% %animal%")
    for i in range(10):
        print(fmt(i))

    fmt.plot_encryption()



import os.path


class OptionGroup(object):
    def __new__(cls, options: str | list, *a, **kw):
        # print(options)
        if isinstance(options, (Charset, Wordlist)):
            return options
        if isinstance(options, str):
            return Charset(options)
        if all(isinstance(o, str) and len(o) == 1 for o in options):
            return Charset("".join(options))
        if isinstance(options, list):
            options = [str(v) for v in options]
        return Optionset(options, *a, **kw)

    def to_int(self, value: str) -> int:
        if value in self:
            return self.index(value)
        n = len(self)
        value = [v for v in value if v in self]
        vals = [self.index(v) * (n ** (len(value) - i - 1)) for i, v in enumerate(value)]
        i = sum(vals)
        return i

    def from_int(self, value: int) -> str | list:
        assert value >= 0, "value must be positive"
        n = len(self)
        if value < n:
            return self[value]
        s = []
        while value:
            s = [self[value % n]] + s
            value //= n
        if all(isinstance(_c, str) for _c in s):
            s = "".join(s)
        return s

    def bit_length(self):
        if len(self) == 0:
            return 0
        return (len(self)-1).bit_length()

    def __add__(self, other):
        return type(self)(super().__add__([v for v in other if v not in self]))

    def __radd__(self, other):
        return type(self)(super().__radd__([v for v in self if v not in other]))

    def __sub__(self, other):
        return type(self)([v for v in self if v not in other])

    def __rsub__(self, other):
        return type(self)([v for v in other if v not in self])


class Optionset(OptionGroup, list):
    def __new__(cls, *a, **kw):
        o = list.__new__(cls)
        o.__init__(*a, **kw)
        return o


class Wordlist(Optionset):
    wordlist_folder = os.path.join(os.path.dirname(__file__), "wordlists")

    def __init__(self, fn, load: bool = True):
        super().__init__()

        if (not os.path.exists(fn)) and os.path.exists(new_fn := os.path.join(self.wordlist_folder, fn)):
            fn = new_fn

        if isinstance(fn, str):
            self.fn = fn
            self.loaded = False
        else:
            self.fn = None
            self.extend(fn)
            self.loaded = True
        if load:
            self.load()

    def load(self):
        if not self.loaded:
            fn = self.fn
            if fn.endswith(".json"):
                import json
                with open(fn) as f:
                    d = json.load(f)
            elif fn.endswith(".yml") or fn.endswith(".yaml"):
                import yaml
                with open(fn) as f:
                    d = yaml.safe_load(f)
            else:
                d = [v.replace('\n', '') for v in open(fn).readlines()]
            self.extend(d)
            self.loaded = True
            return d
        return self


class Charset(OptionGroup, str):
    def __new__(cls, s):
        if isinstance(s, list):
            s = "".join(s)
        o = str.__new__(cls, s)
        o.__init__()
        return o

    def __getitem__(self, item):
        return Charset(''.join(super().__getitem__(item)))

    def from_int(self, i):
        o = super().from_int(i)
        return "".join(o)


class Options(dict):
    wordsets = {
        'colors_16': ['aqua', 'black', 'blue', 'fuchsia', 'gray', 'green', 'lime', 'maroon', 'navy', 'olive', 'purple',
                      'red', 'silver', 'teal', 'white', 'yellow'],
        'singular_animals_128': Wordlist('128_singular_animals.txt', load=False),
        'plural_animals_128': Wordlist('128_plural_animals.txt', load=False),
        'rgb_24bit': [(r, g, b) for b in range(256) for g in range(256) for r in range(256)],
        'singular_nouns_1k': Wordlist('1k_singular_nouns.txt', load=False),
        'plural_nouns_1k': Wordlist('1k_plural_nouns.txt', load=False),
        'adjectives_1k':Wordlist('1k_adjectives.txt', load=False),
    }

    charsets = {
        'binary': '01',
        'octadecimal': '01234567',
        'decimal': '0123456789',
        'hexadecimal': '0123456789abcdef',
        'lowercase': 'abcdefghijklmnopqrstuvwxyz',
        'uppercase': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'punctuation': '!"' + "#$%&'()*+,-./:;<=>?@[\]^_`{|}~'",
        'whitespace': "".join(chr(i) for i in [32, 9, 10, 13, 11, 12]),
        'ascii_128': "".join(chr(i) for i in range(128)),
        'ascii_128_unescaped': "".join(chr(i) for i in range(128) if '\\' not in repr(chr(i))),
        'ascii_256': "".join(chr(i) for i in range(256)),
        'ascii_256_unescaped': "".join(chr(i) for i in range(256) if '\\' not in repr(chr(i))),
        'ascii_512': "".join(chr(i) for i in range(512)),
        'ascii_512_unescaped': "".join(chr(i) for i in range(512) if '\\' not in repr(chr(i))),
        'ascii_1028': "".join(chr(i) for i in range(1028)),
        'ascii_1028_unescaped': "".join(chr(i) for i in range(1028) if '\\' not in repr(chr(i))),
        'greek_lowercase': 'αβγδεζηθικλμνξοπρστυφχψω',
        'greek_uppercase': 'ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ',
        'vowels_lowercase': 'aeiou',
        'vowels_uppercase': 'AEIOU',
        'vowels': 'aeiouAEIOU',
        'consonants_lowercase': 'bcdfghjklmnpqrstvwxyz',
        'consonants_uppercase': 'BCDFGHJKLMNPQRSTVWXYZ',
        'consonants': 'bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ',
    }

    super_charsets = {
        'alphabet': charsets['lowercase'] + charsets['uppercase'],
        'alphanumeric_lowercase': charsets['lowercase'] + charsets['decimal'],
        'alphanumeric_uppercase': charsets['uppercase'] + charsets['decimal'],
        'alphanumeric': charsets['lowercase'] + charsets['uppercase'] + charsets['decimal'],
        'greek_alphabet': charsets['greek_lowercase'] + charsets['greek_uppercase']
    }
    charsets.update(super_charsets)

    groups = wordsets.copy()
    groups.update(charsets)
    
    wordset_aliases = {
        'colors': 'colors_16',
        "color": "colors_16",
        'plural_animal': 'plural_animals_128',
        'plural_animals': 'plural_animals_128',
        'animals': 'plural_animals_128',
        "singular_animal": "singular_animals_128",
        "animal": "singular_animals_128",
        'plural_nouns': 'plural_nouns_1k',
        "plural_noun": "plural_nouns_1k",
        "singular_nouns": "singular_nouns_1k",
        "singular_noun": "singular_nouns_1k",
        "noun": "singular_nouns_1k",
        'adjectives': 'adjectives_1k',
        'adjective': 'adjectives_1k',
    }
    charset_aliases = {
        'ascii_lowercase': 'lowercase',
        'az': 'lowercase',
        'ascii_uppercase': 'uppercase',
        'AZ': 'uppercase',
        'ALPHABET': 'uppercase',
        'aZ': 'alphabet',
        'english_alphabet': 'alphabet',
        'az9': 'alphanumeric_lowercase',
        'AZ9': 'alphanumeric_uppercase',
        'aZ9': 'alphanumeric',
        '09': 'decimal',
        'digits': 'decimal',
        'digit': 'decimal',
        '07': 'octadecimal',
        'octdigits': 'octadecimal',
        'octdigit': 'octadecimal',
        'hexdigits': 'hexadecimal',
        'hexdigit': 'hexadecimal',
        'hex': 'hexadecimal',
        'hs': 'hexadecimal',
        'bin': 'binary',
        'bs': 'binary',
        'printable': 'ascii_128_unescaped',
        'greek': 'greek_alphabet',
        'GREEK': 'greek_uppercase',
        'VOWELS': 'vowels_uppercase',
        'VOWEL': 'vowels_uppercase',
        'CONSONANTS': 'consonants_uppercase',
        'CONSONANT': 'consonants_uppercase',
        "vowel": "vowels",
        "consonant": "consonants"
    }
    aliases = wordset_aliases.copy()
    aliases.update(charset_aliases)

    def __init__(self, groups: dict = None, **group_kwargs):
        if groups is not None:
            self.groups.update(groups)
        self.groups.update(group_kwargs)
        super().__init__(self.groups)
      
    def __getitem__(self, item):
        if isinstance(item, str) and item.isnumeric():
            item = int(item)
        if isinstance(item, int):
            v = [str(v) for v in range(item)]
        elif isinstance(item, slice):
            start = item.start if item.start is not None else 0
            step = item.step if item.step is not None else 1
            v = [str(v) for v in range(start, item.stop, step)]
        elif item in self:
            v = super().__getitem__(item)
        elif item in self.aliases:
            v = self[self.aliases[item]]
        else:
            return None
        # print(f"{v=}, {type(v)}")
        if not isinstance(v, OptionGroup):
            v = OptionGroup(v)
            self[item] = v
        if isinstance(v, Wordlist) and not v.loaded:
            v.load()
        # print(f"{v=}")
        return v
    
    def __dir__(self):
        return list(self.keys()) + list(self.aliases.keys())

    def __repr__(self):
        return 'charsets and wordlists'

    @staticmethod
    def trim(s):
        if isinstance(s, str):
            if s.startswith('0x'):
                s = s[2:]
            if s[0] == 'b' and all(v in '01 ' for v in s[1:]):
                s = s[1:]
                s = s.replace(' ', '')
        return s
    
    @classmethod
    def identify(cls, s):
        s = cls.trim(s)

        for group_name, group in cls.groups.items():
            s_in_v = isinstance(s, str) and isinstance(group, str) and all(_char in group for _char in s)
            s_in_v |= isinstance(group, list) and s in group

            try:
                if s_in_v:
                    if group_name in cls.aliases.values():
                        for alias, val in cls.aliases.items():
                            if val == group_name:
                                return alias
                    return group_name
            except:
                pass
        return -1


options = Options()

if __name__ == "__main__":
    o = options
    # print(o.animals)

